# -*- coding: utf-8 -*-

from __future__ import annotations

import re
from random import choices
from string import ascii_letters
from string import digits
from string import punctuation
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional


def random_string(
    vocabulary: Optional[str] = None,
    length: int = 20,
    exclude: Optional[str] = None,
) -> str:
    """Create random string"""

    if not vocabulary:
        vocabulary = ascii_letters + digits + punctuation

    choices_: List[str] = []
    while len(choices_) < length:
        char = choices(vocabulary)[0]
        if not exclude or (exclude and char not in exclude):
            choices_.append(char)

    return "".join(choices_)


def get_batches(list_: List, n: int) -> Iterator:
    """
    Divide a list into chunks...

    :param list_: List to divide.
    :param n: Number of elements by chunk.
    :return:
    """

    for i in range(0, len(list_), n):
        yield list_[i : i + n]


def remove_attributes(record: Dict, attrs: List[str]) -> None:
    """Remove attributes from a Dict"""

    for key in attrs:
        if key in record:
            del record[key]


def rename_attributes(record: Dict, mapper: Dict[str, str]) -> None:
    """Rename the object attributes using the column_mapper"""

    _items = list(record.items())
    for key, value in _items:
        if key in mapper:
            record[mapper[key]] = value
            del record[key]


def add_attributes(record: Dict, expected_attrs: List[str]):
    """
    The function add to the dictionary the list of
    attributes (if not exists) expected
    using None...

    :param record: Dictionary to update.
    :param expected_attrs: Attributes to add.
    :return:
    """

    for key in expected_attrs:
        if key not in record:
            record[key] = None


def convert_data_type(
    data: Dict[str, str],
    columns_type_mapper: Dict[str, Any],
) -> None:
    """
    Update (cast) the value depending on the specified type...

    :param data: Dictionary to update.
    :param columns_type_mapper: Dictionary that specify the data types.
    """

    if not columns_type_mapper:
        return

    # Safe type mapping
    type_converters = {
        "str": str,
        "int": int,
        "float": float,
        "bool": bool,
        "list": list,
        "dict": dict,
    }

    for key, type_name in columns_type_mapper.items():
        if key in data and data[key] is not None:
            if type_name not in type_converters:
                raise ValueError(f"Unsupported type: {type_name}")

            try:
                value = data[key].strip() if type_name == "float" else data[key]
                converter = type_converters[type_name]
                data[key] = converter(value)

            except Exception as e:
                raise e.__class__(f"Cannot convert {value} to {type_name}: {e}")


def to_snake_case(string: str) -> str:
    """It converts a string from camel-case to snake-case"""

    return "".join(["_" + i.lower() if i.isupper() else i for i in string]).lstrip("_")


def flatten_json(data: Dict, flatten_sublist: bool = False):
    """
    Utility function for flattening dictionary objects...

    :param data: Object to flatten.
    :param flatten_sublist: Set as True if you want to flatten sublist objects.

    :return: Flatten data (dictionary).
    """

    res = {}

    def flatten(x, name: str = ""):
        if isinstance(x, dict):
            for a in x:
                flatten(x[a], name + a + "_")

        elif flatten_sublist and isinstance(x, list):
            i = 0
            for a in x:
                flatten(a, name + str(i) + "_")
                i += 1

        else:
            res[name[:-1]] = x

    flatten(data)
    return res


def to_one_line(multiline_string: str) -> str:
    """It converts a multiline string to a single line one"""

    return re.sub(
        pattern=r"\s+", repl=" ", string=multiline_string.replace("\n", "")
    ).strip()


def bytes_to_str(data: List | Dict | bytes, encoding: str = "utf-8"):
    """
    Convert bytes to str in keys and values...

    Example:
        {b"key": b"value"})         -> {"key": "value"}
        {b"key": [1, 2, b"3"]}      -> {"key": [1, 2, "3"]}
        [1, 2, b"3", {b"a": b"a"}]  -> [1, 2, "3", {"a": "a"}]
        b"test")                    -> "test"
    """

    def convert(obj_, encoding_, result=None):
        if not result:
            result = {}

        if isinstance(obj_, dict):
            dic_ = {}
            for key in obj_:
                previous_key = key
                if isinstance(key, bytes):
                    key = key.decode()

                dic_[key] = convert(obj_[previous_key], encoding_, dic_)

            return dic_

        elif isinstance(obj_, list):
            return [convert(x, encoding_, result or {}) for x in obj_]

        else:
            return obj_.decode() if isinstance(obj_, bytes) else obj_

    return convert(data, encoding)



# -*- coding: utf-8 -*-

import asyncio
import inspect
import threading
from functools import wraps


class SyncWrapper:
    """
    It provides an interface (mechanism) to execute async code from
    sync apps via a background thread that keeps a single event
    loop alive, avoiding `asyncio.run()` which creates and
    destroys an event loop each time and becomes
    expensive for many method calls.

    .. code-block:: python

        class Test:
            def sync_method(self) -> str:
                return self.__class__.__name__

            async def testing(self) -> bool:
                await sleep(0)
                return True

        sync_instance = SyncWrapper(instance)
        assert Test.__name__, sync_instance.sync_method()
        assert sync_instance.testing(), True
        sync_instance.close()
    ..
    """

    def __init__(self, async_instance):
        self._async_instance = async_instance
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._thread.start()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __getattr__(self, name):
        attr = getattr(self._async_instance, name)
        if inspect.iscoroutinefunction(attr):

            @wraps(attr)
            def sync_method(*args, **kwargs):
                coro = attr(*args, **kwargs)
                future = asyncio.run_coroutine_threadsafe(coro, self._loop)
                return future.result()

            return sync_method

        return attr

    def _run_loop(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def close(self):
        if hasattr(self._async_instance, "close") and inspect.iscoroutinefunction(self._async_instance.close):
            fut = asyncio.run_coroutine_threadsafe(self._async_instance.close(), self._loop)
            fut.result()

        self._loop.call_soon_threadsafe(self._loop.stop)
        self._thread.join()
        self._loop.close()



# -*- coding: utf-8 -*-

"""
This module provides a set of functions that could
be useful in files and data manipulation.
"""

import csv

from io import StringIO
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional


def records_from_csv(
    path: str,
    file_args: Optional[Dict] = None,
    reader_args: Optional[Dict] = None,
    encoding: Optional[str] = "utf-8",
) -> Iterator[Dict]:
    """
    Returns records from a CSV file...

    :return: An iterable of dictionaries.
    :rtype: Iterator[Dict]
    """

    file_args = file_args or {}
    reader_args = reader_args or {}

    with open(path, encoding=encoding, **file_args) as f_:
        yield from csv.DictReader(f_, **reader_args)


def records_from_buffer(
    data: str,
    newline: Optional[str] = None,
    **kwargs,
) -> Iterator[Dict]:
    """
    Returns records from file-like object...

    :return: An iterable of dictionaries.
    :rtype: Iterator[Dict]
    """

    dialect = kwargs.pop("dialect", "excel")
    quoting = kwargs.pop("quoting", csv.QUOTE_NONNUMERIC)

    yield from csv.DictReader(
        StringIO(data, newline=newline),
        dialect=dialect,
        quoting=quoting,
        **kwargs,
    )


def records_to_buffer(
    records: List[Dict],
    columns: Optional[List[str]] = None,
    with_headers: bool = True,
    **kwargs,
) -> StringIO:
    """
    Save records from file-like object...

    :param records: A collection of dictionary records to write.
    :type records: Iterable[Dict]
    :param columns: A list of column names specifying the order of fields.
    :type columns: List[str]
    :param with_headers: Whether to include column headers in the CSV (default is True).
    :type with_headers: bool, optional

    :return: The buffer.
    :rtype: StringIO
    """

    buffer = StringIO()
    dialect = kwargs.pop("dialect", "excel")
    quoting = kwargs.pop("quoting", csv.QUOTE_NONNUMERIC)

    if not columns:
        if not records:
            raise ValueError(
                "Cannot determine columns: 'columns' parameter is None "
                "and 'records' is empty."
            )

        columns = list(records[0].keys())

    csv_writer = csv.DictWriter(
        buffer,
        fieldnames=columns,
        dialect=dialect,
        quoting=quoting,
        **kwargs,
    )

    if with_headers:
        csv_writer.writeheader()

    csv_writer.writerows(records)
    return buffer


def records_to_csv(  # pylint: disable=too-many-positional-arguments
    file_path: str,
    records: Iterable[Dict],
    columns: List[str],
    with_headers: bool = True,
    quoting: int = csv.QUOTE_ALL,
    encoding: Optional[str] = "utf-8",
    dialect=csv.excel,
    file_args: Optional[Dict] = None,
    writer_args: Optional[Dict] = None,
) -> str:
    """
    It writes records (dictionaries) to a CSV file...

    :param file_path: The path where the CSV file will be saved.
    :type file_path: str
    :param records: A collection of dictionary records to write.
    :type records: Iterable[Dict]

    :param columns: A list of column names specifying the order of fields.
    :type columns: List[str]
    :param with_headers: Whether to include column headers in the CSV (default is True).
    :type with_headers: bool, optional

    :param quoting: Quoting style for the CSV (default is csv.QUOTE_ALL).
    :type quoting: int, optional

    :param encoding: The encoding to use.
    :type encoding: str

    :param dialect: The CSV dialect to use (default is csv.excel).
    :type dialect: csv.Dialect, optional

    :param file_args: Additional arguments for `open()` when writing the file (default is None).
    :type file_args: Dict, optional
    :param writer_args: Additional arguments for `csv.DictWriter` (default is None).
    :type writer_args: Dict, optional

    :return: The file path of the saved CSV.
    :rtype: str

    :raises ValueError: If `columns` is empty or records contain keys not in `columns`.
    :raises IOError: If there is an error writing to the file.
    """

    file_args = file_args or {}
    writer_args = writer_args or {}

    with open(file_path, mode="w", encoding=encoding, **file_args) as file:
        writer = csv.DictWriter(
            file,
            fieldnames=columns,
            dialect=dialect,
            quoting=quoting,  # type: ignore[arg-type]
            **writer_args,
        )

        if with_headers:
            writer.writeheader()

        writer.writerows(records)

    return file_path



# -*- coding: utf-8 -*-

from functools import update_wrapper, wraps
from typing import Callable


class CountCalls:
    """It provides the mechanism to count invocations -- class decorator"""

    def __init__(self, fcn: Callable):
        update_wrapper(self, fcn)
        self.calls_number = 0
        self.fcn = fcn

    def __call__(self, *args, **kwargs):
        self.calls_number += 1
        return self.fcn(*args, **kwargs)


def count_calls(fcn: Callable) -> Callable:
    """
    It provides the mechanism to count invocations...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :return: The wrapped function.
    :rtype: Callable
    """

    @wraps(fcn)
    def wrapper_count_calls(*args, **kwargs):
        wrapper_count_calls.calls_number += 1
        return fcn(*args, **kwargs)

    wrapper_count_calls.calls_number = 0  # type: ignore[attr-defined]
    return wrapper_count_calls



# -*- coding: utf-8 -*-

try:
    from enum import StrEnum as _StrEnum

except ImportError:
    from .compatibility import StrEnum as _StrEnum  # type: ignore[assignment]

from .compatibility import Self

# Type alias that works with both standard library
# and the custom HTTPStatus...
StrEnum = _StrEnum

__all__ = [
    "Self",
    "StrEnum",
]



# -*- coding: utf-8 -*-

from __future__ import annotations

from multiprocessing.pool import Pool
from typing import List, Optional

from core_mixins.decorators import timer
from core_mixins.interfaces.task import ITask
from core_mixins.interfaces.task import TaskResult
from core_mixins.interfaces.task import TaskStatus


class TasksManager:
    """It manages the execution for the registered tasks"""

    def __init__(self, tasks: List[ITask]):
        self.tasks = tasks

    def execute(
        self,
        task_name: Optional[str] = None,
        parallelize: Optional[bool] = False,
        processes: Optional[int] = None,
    ) -> Optional[List[TaskResult] | TaskResult]:
        """
        Execute all registered tasks. An exception in one task should not
        stop the execution of the others...

        Example of results:

        .. code-block:: python

            [
                TaskResult(status=TaskStatus.SUCCESS, result=...),
                TaskResult(status=TaskStatus.ERROR, error=...)
            ]
        ..

        :param task_name: If defined, only that specific task will be executed.
        :type task_name: str
        :param parallelize: It defines if you want to execute the tasks in parallel.
        :type parallelize: bool
        :param processes: Number of parallel process.
        :type processes: int

        :return: The list of the execution results.
        :rtype: List[TaskResult]
        """

        if not self.tasks:
            return None

        if task_name:
            for task in self.tasks:
                if task_name == task.name:
                    return self._execute(task)

            raise Exception(f"Task [{task_name}] is not registered!")

        res = []
        if not parallelize:
            for task in self.tasks:
                res.append(self._execute(task))

        else:
            with Pool(processes=processes) as pool:
                res = pool.map(self._execute, self.tasks)

        return res

    @staticmethod
    def _execute(task: ITask) -> TaskResult:
        try:

            @timer
            def _execute():
                return task.execute()

            res, seconds = _execute()

            return TaskResult(
                status=TaskStatus.SUCCESS,
                execution_time=seconds,
                result=res,
            )

        except Exception as error:
            return TaskResult(
                status=TaskStatus.ERROR,
                error=error,
            )



# -*- coding: utf-8 -*-

"""
Base interface for classes that acts like factories.
"""

import threading
from abc import ABC, abstractmethod
from inspect import isabstract
from typing import Type, Dict, Set, Optional

from core_mixins.compatibility import Self


class IFactory(ABC):
    """
    Base interface for classes that acts like
    factories. The IFactory pattern
    allows you to:

      - Automatically register concrete implementations.
      - Discover available implementations at runtime.
      - Create instances by string reference.
      - Maintain separate registries for different factory types.
    """

    _impls: Dict[str, Type[Self]] = {}
    _lock = threading.Lock()

    def __init_subclass__(cls, **kwargs) -> None:
        """
        Automatically register concrete subclasses in both the global and
        per-abstract-class registries. Each abstract subclass gets its own
        _impls dict.
        """
        super().__init_subclass__(**kwargs)

        if isabstract(cls):
            # Each abstract (new base) class will have its own
            # register. This way each new base interface does not need
            # to redefine `_impls`...
            cls._impls = {}

        else:
            with IFactory._lock:
                for base in cls.__mro__:
                    if (
                        issubclass(base, IFactory)
                        and isabstract(base)
                        and hasattr(base, "_impls")
                    ):
                        key = cls.registration_key()
                        if key in base._impls:
                            raise ValueError(
                                f"Registration key '{key}' already exists "
                                f"in {base.__name__} registry. "
                                f"Existing class: {base._impls[key].__name__}, "
                                f"Attempting to register: {cls.__name__}"
                            )

                        base._impls[key] = cls

    @classmethod
    @abstractmethod
    def registration_key(cls) -> str:
        """
        It returns the name (reference) for the key used
        to register, like: return `self.__name__`
        """

    @classmethod
    def get_registered_classes(cls) -> Set[str]:
        """It returns the current registered implementations"""
        return set(cls._impls)

    @classmethod
    def get_class(cls, cls_reference: str) -> Optional[Type[Self]]:
        """It returns the reference (class type) by reference"""
        return cls._impls.get(cls_reference, None)



# -*- coding: utf-8 -*-

from .utils import (
    records_from_buffer,
    records_from_csv,
    records_to_buffer,
    records_to_csv,
)


__all__ = [
    "records_from_buffer",
    "records_from_csv",
    "records_to_buffer",
    "records_to_csv",
]



# -*- coding: utf-8 -*-

from .async_ import SyncWrapper
from .cache import cache
from .count_calls import CountCalls, count_calls
from .repeat import repeat
from .retry import retry
from .singleton import singleton
from .timer import timer


__all__ = [
    "cache",
    "CountCalls",
    "count_calls",
    "repeat",
    "retry",
    "singleton",
    "SyncWrapper",
    "timer",
]



# -*- coding: utf-8 -*-

import sys
import traceback

from typing import Tuple, Dict, List


def get_exception_data() -> Tuple[str, str, List[Dict]]:
    """
    To retrieve the error information and the stack trace...

    Example:
        try:
            8/0

        except ZeroDivisionError as error:
            type_, message, trace = get_exception_data()

        Exception type: ZeroDivisionError
        Exception message: division by zero
        Stack trace: ...

    :return:
        The error information in a tuple like:
        (exception_type, exception_message, stack_trace)
    :rtype: Tuple[str, str, List[Dict]]
    """

    ex_type, ex_value, ex_traceback = sys.exc_info()

    stack_trace = [
        {
            "File": trace.filename,
            "Line Number": trace.lineno,
            "Line": trace.line,
            "Function": trace.name,
        }
        for trace in traceback.extract_tb(ex_traceback)
    ]

    return ex_type.__name__ if ex_type else "Unknown", str(ex_value), stack_trace



# -*- coding: utf-8 -*-

"""
This module provides a base implementation for ETL tasks
under the project ecosystem.
"""

from abc import ABC, abstractmethod
from enum import Enum
from logging import Logger
from typing import Any, Optional

from core_mixins.interfaces.factory import IFactory


class TaskStatus(str, Enum):
    """Possible status a task can have during its life"""

    CREATED = "CREATED"
    EXECUTING = "EXECUTING"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"


class ITask(IFactory, ABC):
    """Base implementations for different tasks/processes"""

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        logger: Optional[Logger] = None,
    ) -> None:
        self._name = name
        self.description = description
        self._status = TaskStatus.CREATED
        self.logger = logger

    @classmethod
    def registration_key(cls) -> str:
        """It returns the key used to register the class"""
        return cls.__name__

    @property
    def name(self):
        """It returns the task identification"""
        return self._name or self.registration_key()

    @property
    def status(self) -> TaskStatus:
        """It returns the current status of the task"""
        return self._status

    @status.setter
    def status(self, status: TaskStatus) -> None:
        self._status = status

    @abstractmethod
    def execute(self, *args, **kwargs) -> Any:
        """You must implement the task's process"""

    def info(self, message) -> None:
        """Log entry with severity 'INFO'"""

        if self.logger:
            self.logger.info(f"{self.name} | {message}")

    def warning(self, message) -> None:
        """Log entry with severity 'WARNING'"""

        if self.logger:
            self.logger.warning(f"{self.name} | {message}")

    def error(self, error) -> None:
        """Log entry with severity 'ERROR'"""

        if self.logger:
            self.logger.error(f"{self.name} | {error}")


class TaskResult:
    """
    Represents the result of a task execution that
    exposes `status`, `result`, or the error.
    """

    def __init__(
        self,
        status: TaskStatus,
        result: Optional[Any] = None,
        error: Optional[Exception] = None,
        execution_time: float = 0.0,
    ) -> None:
        self.status = status
        self.execution_time = execution_time
        self.result = result
        self.error = error

    def __repr__(self) -> str:
        if self.status == TaskStatus.SUCCESS:
            return f"TaskResult(status={self.status}, result={self.result})"

        elif self.status == TaskStatus.ERROR:
            return f"TaskResult(status={self.status}, error={self.error})"

        else:
            return f"TaskResult(status={self.status})"

    @property
    def is_success(self) -> bool:
        """Check if the task execution was successful"""
        return self.status == TaskStatus.SUCCESS

    @property
    def is_error(self) -> bool:
        """Check if the task execution failed"""
        return self.status == TaskStatus.ERROR


class TaskException(Exception):
    """Custom exception for Tasks"""



# -*- coding: utf-8 -*-

from functools import wraps
from typing import Callable, Optional


def repeat(fcn: Optional[Callable] = None, *, times: int = 2) -> Callable:
    """
    Repeat n times the function and return the list of returned values...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :param times: Number of times the function will be invoked.
    :type times: int

    :return: The wrapped function.
    :rtype: Callable
    """

    def decorator_repeat(func):
        @wraps(func)
        def wrapper_repeat(*args, **kwargs):
            values = []
            for _ in range(times):
                values.append(func(*args, **kwargs))

            return values

        return wrapper_repeat

    if not fcn:
        return decorator_repeat

    return decorator_repeat(fcn)



# -*- coding: utf-8 -*-

from functools import wraps
from logging import Logger
from time import sleep
from typing import Callable, Type, Tuple, Optional


def retry(
    fcn: Optional[Callable] = None,
    tries: Optional[int] = 3,
    delay: Optional[int] = 1,
    backoff: Optional[int] = 2,
    exceptions: Optional[Tuple[Type[BaseException]]] = (Exception,),
    logger: Optional[Logger] = None,
) -> Callable:
    """
    It retries the decorated function using an exponential
    backoff in case of errors (exceptions) in the
    execution...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :param tries: Number of retries.
    :type tries: int

    :param delay: Delay in seconds between invocations.
    :type delay: int
    :param backoff: Exponential backoff to used between retries.
    :type backoff: int

    :param exceptions:
        Exceptions to capture for the retries. If the raised exception is not
        here, no retries are performed.
    :type exceptions: Callable

    :param logger: The function being decorated.
    :type logger: Callable


    :return: The wrapped function.
    :rtype: Callable
    """

    def decorator_retry(_fcn: Callable) -> Callable:
        @wraps(_fcn)
        def function_retry(*args, **kwargs):
            _tries, _delay = tries, delay
            while _tries > 1:
                try:
                    return _fcn(*args, **kwargs)

                except exceptions as error:
                    if logger:
                        logger.warning(f"Retrying in {_delay}. Because: {error}")

                    sleep(_delay)
                    _tries -= 1
                    _delay *= backoff

            return _fcn(*args, **kwargs)

        return function_retry

    if not fcn:
        return decorator_retry

    return decorator_retry(fcn)



# -*- coding: utf-8 -*-

from functools import wraps
from time import perf_counter
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple


def timer(fcn: Optional[Callable] = None) -> Callable:
    """
    Using this decorator you will get a tuple in the form
    of: result, execution_time...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :return: The wrapped function.
    :rtype: Callable
    """

    def decorator_timer(func: Callable):
        @wraps(func)
        def wrapper_timer(*args, **kwargs) -> Tuple[Any, float]:
            start_time, res = perf_counter(), func(*args, **kwargs)
            return res, (perf_counter() - start_time)

        return wrapper_timer

    if not fcn:
        return decorator_timer

    return decorator_timer(fcn)



# -*- coding: utf-8 -*-

from functools import wraps
from typing import Callable, Type


def singleton(cls: Type) -> Callable:
    """
    Make a class a Singleton class (only one instance)

    :param cls: The class being decorated.
    :type cls: ``Type``
    :return: The wrapped function.
    :rtype: Callable
    """

    @wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)

        return wrapper_singleton.instance

    wrapper_singleton.instance = None  # type: ignore[attr-defined]
    return wrapper_singleton



# -*- coding: utf-8 -*-

from enum import Enum

try:
    from typing import Self

except ImportError:
    from typing_extensions import Self  # type: ignore[assignment]


class StrEnum(str, Enum):
    """
    For backward compatibility because StrEnum class
    was added in Python 3.11...
    """

    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, str):
            raise TypeError(f"{value} is not a string")

        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj



# -*- coding: utf-8 -*-

from functools import wraps
from typing import Callable


def cache(fcn: Callable) -> Callable:
    """
    It maintains a cache of previous function call results that can be
    used to optimize the performance...

    :param fcn: The function being decorated.
    :type fcn: Callable
    :return: The wrapped function.
    :rtype: Callable
    """

    @wraps(fcn)
    def wrapper_cache(*args, **kwargs):
        cache_key = args + tuple(kwargs.items())
        if cache_key not in wrapper_cache.cache:
            wrapper_cache.cache[cache_key] = fcn(*args, **kwargs)

        return wrapper_cache.cache[cache_key]

    wrapper_cache.cache = {}  # type: ignore[attr-defined]
    return wrapper_cache
